import requests

from django.http import HttpResponseForbidden

from .forms import PaymentForm
from .. import get_payment_model, BasicProvider

Payment = get_payment_model()


class AuthorizeNetProvider(BasicProvider):

    def __init__(self, *args, **kwargs):
        self.login_id = kwargs.pop('login_id')
        self.transaction_key = kwargs.pop('transaction_key')
        self.endpoint = kwargs.pop('endpoint',
                                   'https://test.authorize.net/gateway/transact.dll')
        super(AuthorizeNetProvider, self).__init__(*args, **kwargs)

    def get_transactions_data(self):
        data = {
            'x_amount': self.payment.total,
            'x_currency_code': self.payment.currency,
            'x_description': self.payment.description,
            'x_first_name': self.payment.billing_first_name,
            'x_last_name': self.payment.billing_last_name,
            'x_address': u"%s, %s" % (self.payment.billing_address_1,
                                      self.payment.billing_address_2),
            'x_city': self.payment.billing_city,
            'x_zip': self.payment.billing_postcode,
            'x_country': self.payment.billing_country_area
        }
        return data

    def get_product_data(self, extra_data=None):
        data = self.get_transactions_data()

        if extra_data:
            data.update(extra_data)

        data.update({
            'x_login': self.login_id,
            'x_tran_key': self.transaction_key,
            'x_delim_data': True,
            'x_delim_char': "|",
            'x_method': "CC",
            'x_type': "AUTH_CAPTURE"})

        return data

    def get_payment_response(self, extra_data=None):
        post = self.get_product_data(extra_data)
        return requests.post(self.endpoint, data=post)

    def get_form(self, data=None):
        return PaymentForm(data=data, payment=self.payment, provider=self, action='')

    def process_data(self, request):
        return HttpResponseForbidden('FAILED')



from datetime import date
from calendar import monthrange
import re

from django import forms
from django.core import validators
from django.utils.translation import ugettext_lazy as _

from . import widgets

CARD_TYPES = {
    'visa': ('^4[0-9]{12}(?:[0-9]{3})?$', 'VISA'),
    'mastercard': ('^5[1-5][0-9]{14}$', 'MasterCard'),
    'discover': ('^6(?:011|5[0-9]{2})[0-9]{12}$', 'Discover'),
    'amex': ('^3[47][0-9]{13}$', 'American Express'),
    'jcb': ('^(?:(?:2131|1800|35\d{3})\d{11})$', 'JCB'),
    'diners': ('^(?:3(?:0[0-5]|[68][0-9])[0-9]{11})$', 'Diners Club')}


class CreditCard(object):

    def __init__(self, number):
        self.number = number
        self.card_type, self.card_type_fullname = self.get_credit_card_type(number)

    def __repr__(self):
        return 'CreditCard(number=%s, card_type=%s, card_type_fullname=%s)' % \
               (self.number, self.card_type, self.card_type_fullname)

    def __unicode__(self):
        return self.number

    def get_credit_card_type(self, number):
        for card_type, details in CARD_TYPES.items():
            regexp, name = details
            if re.match(regexp, number):
                return card_type, name
        return None, None


class CreditCardNumberField(forms.CharField):

    widget = widgets.CreditCardNumberWidget
    default_error_messages = {
        'invalid': _(u'Please enter a valid card number'),
        'invalid_type': _(u'We accept only %(valid_types)s')}

    def __init__(self, valid_types=None, *args, **kwargs):
        self.valid_types = valid_types
        kwargs['max_length'] = kwargs.pop('max_length', 32)
        super(CreditCardNumberField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        cleaned = re.sub('[\s-]', '', value)
        if value and not cleaned:
            raise forms.ValidationError(self.error_messages['invalid'])
        return CreditCard(number=cleaned)

    def validate(self, card):
        value = card.number
        if value in validators.EMPTY_VALUES and self.required:
            raise forms.ValidationError(self.error_messages['required'])
        if value and not self.cart_number_checksum_validation(value):
            raise forms.ValidationError(self.error_messages['invalid'])
        if value and not self.valid_types is None \
                and not card.card_type in self.valid_types:
            raise forms.ValidationError(self.error_messages['invalid_type'] %
                                        {'valid_types': ', '.join(map(self.get_card_type_fullname,
                                                                      self.valid_types))})

    def run_validators(self, card):
        super(CreditCardNumberField, self).run_validators(card.number)

    def get_card_type_fullname(self, card_type):
            fullname = ''
            type_detail = CARD_TYPES.get(card_type)
            if type_detail:
                _, fullname = type_detail
            return fullname

    def cart_number_checksum_validation(self, number):
        digits = []
        even = False
        if not number.isdigit():
            return False
        for digit in reversed(number):
            digit = ord(digit) - ord('0')
            if even:
                digit = digit * 2
                if digit >= 10:
                    digit = digit % 10 + digit / 10
            digits.append(digit)
            even = not even
        return sum(digits) % 10 == 0 if digits else False


# Credit Card Expiry Fields from:
# http://www.djangosnippets.org/snippets/907/
class CreditCardExpiryWidget(forms.MultiWidget):
    """MultiWidget for representing credit card expiry date."""
    def decompress(self, value):
        if value:
            return [value.month, value.year]
        else:
            return [None, None]

    def format_output(self, rendered_widgets):
        html = u' / '.join(rendered_widgets)
        return u'<span style="white-space: nowrap">%s</span>' % html


# From https://github.com/zen4ever/django-authorizenet
class CreditCardExpiryField(forms.MultiValueField):

    EXP_MONTH = [(x, "%02d" % x) for x in xrange(1, 13)]
    EXP_YEAR = [(x, x) for x in xrange(date.today().year,
                                       date.today().year + 15)]

    default_error_messages = {
        'invalid_month': u'Enter a valid month.',
        'invalid_year': u'Enter a valid year.'}

    def __init__(self, *args, **kwargs):
        errors = self.default_error_messages.copy()
        if 'error_messages' in kwargs:
            errors.update(kwargs['error_messages'])

        fields = (
            forms.ChoiceField(
                choices=self.EXP_MONTH,
                error_messages={'invalid': errors['invalid_month']}),
            forms.ChoiceField(
                choices=self.EXP_YEAR,
                error_messages={'invalid': errors['invalid_year']}),
        )

        super(CreditCardExpiryField, self).__init__(fields, *args, **kwargs)
        self.widget = CreditCardExpiryWidget(widgets=[fields[0].widget,
                                                      fields[1].widget])

    def clean(self, value):
        exp = super(CreditCardExpiryField, self).clean(value)
        if date.today() > exp:
            raise forms.ValidationError(
                "The expiration date you entered is in the past.")
        return exp

    def compress(self, data_list):
        if data_list:
            if data_list[1] in forms.fields.EMPTY_VALUES:
                error = self.error_messages['invalid_year']
                raise forms.ValidationError(error)
            if data_list[0] in forms.fields.EMPTY_VALUES:
                error = self.error_messages['invalid_month']
                raise forms.ValidationError(error)
            year = int(data_list[1])
            month = int(data_list[0])
            # find last day of the month
            day = monthrange(year, month)[1]
            return date(year, month, day)
        return None


class CreditCardVerificationField(forms.CharField):

    default_error_messages = {
        'invalid': _(u'Enter a valid security number.')}

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = kwargs.pop('max_length', 4)
        super(CreditCardVerificationField, self).__init__(*args, **kwargs)

    def validate(self, value):
        if value in validators.EMPTY_VALUES and self.required:
            raise forms.ValidationError(self.error_messages['required'])
        if value and not re.match('^[0-9]{3,4}$', value):
            raise forms.ValidationError(self.error_messages['invalid'])



from functools import wraps
from datetime import timedelta
import requests

from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from django.utils import simplejson, timezone

from .forms import PaymentForm
from .. import BasicProvider, RedirectNeeded, get_payment_model

Payment = get_payment_model()


class UnauthorizedRequest(Exception):
    pass


def authorize(fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        self = args[0]
        self.access_token = self.get_access_token()
        response = fun(*args, **kwargs)
        if response.status_code == 401:
            extra_data = (simplejson.loads(self.payment.extra_data)
                          if self.payment.extra_data else {})
            if 'access_token' in extra_data:
                del extra_data['access_token']
                self.payment.extra_data = simplejson.dumps(extra_data)
            self.access_token = self.get_access_token()
            response = fun(*args, **kwargs)
        return response

    return wrapper


class PaypalProvider(BasicProvider):
    '''
    paypal.com payment provider
    '''
    def __init__(self, *args, **kwargs):
        self.secret = kwargs.pop('secret')
        self.client_id = kwargs.pop('client_id')
        self.endpoint = kwargs.pop('endpoint',
                                   'https://api.sandbox.paypal.com')
        self.oauth2_url = self.endpoint + '/v1/oauth2/token'
        self.payments_url = self.endpoint + '/v1/payments/payment'
        self.payment_execute_url = self.payments_url + '/%(id)s/execute/'
        super(PaypalProvider, self).__init__(*args, **kwargs)

    def get_access_token(self):
        extra_data = (simplejson.loads(self.payment.extra_data)
                      if self.payment.extra_data else {})
        created = self.payment.created
        now = timezone.now()
        if ('access_token' in extra_data and
                'expires_in' in extra_data and
                (created + timedelta(seconds=extra_data['expires_in'])) > now):
            return u'%s %s' % (extra_data['token_type'],
                               extra_data['access_token'])
        else:
            headers = {'Accept': 'application/json',
                       'Accept-Language': 'en_US'}
            post = {'grant_type': 'client_credentials'}
            response = requests.post(self.oauth2_url, data=post,
                                     headers=headers,
                                     auth=(self.client_id, self.secret))
            response.raise_for_status()
            data = response.json()
            extra_data.update(data)
            self.payment.extra_data = simplejson.dumps(extra_data)
            return u'%s %s' % (data['token_type'], data['access_token'])

    def get_link(self, name, data):
        try:
            links = filter(lambda url: url['rel'] == name, data['links'])
        except KeyError:
            return None
        return links[0]['href']

    def get_transactions_data(self):
        items = list(self.payment.get_purchased_items())
        sub_total = self.payment.total - self.payment.delivery
        data = {'intent': 'sale',
                'transactions': [{
                'amount': {
                    'total': self.payment.total,
                    'currency': self.payment.currency,
                    'details': {
                        'subtotal': sub_total,
                        'tax': self.payment.tax,
                        'shipping': self.payment.delivery}},
                'item_list': {'items': items},
                'description': self.payment.description}]}
        return data

    def get_product_data(self, extra_data=None):
        return_url = self.get_return_url()
        data = self.get_transactions_data()
        data['redirect_urls'] = {'return_url': return_url,
                                 'cancel_url': return_url}
        data['payer'] = {'payment_method': 'paypal'}
        return data

    @authorize
    def get_payment_response(self, extra_data=None):
        headers = {'Content-Type': 'application/json',
                   'Authorization': self.access_token}
        post = simplejson.dumps(
            self.get_product_data(extra_data))
        return requests.post(self.payments_url, data=post, headers=headers)

    @authorize
    def get_payment_execute_response(self, payer_id):
        headers = {'Content-Type': 'application/json',
                   'Authorization': self.access_token}
        post = {'payer_id': payer_id}
        transaction_id = self.payment.transaction_id
        execute_url = self.payment_execute_url % {'id': transaction_id}
        return requests.post(execute_url, data=simplejson.dumps(post),
                             headers=headers)

    def get_form(self, data=None):
        if not self.payment.id:
            self.payment.save()
        extra_data = (simplejson.loads(self.payment.extra_data)
                      if self.payment.extra_data else {})
        redirect_to = self.get_link('approval_url', extra_data)
        if not redirect_to:
            response = self.get_payment_response()
            response.raise_for_status()
            response_data = response.json()
            redirect_to = self.get_link('approval_url', response_data)
            self.payment.transaction_id = response_data['id']
            extra_data['links'] = response_data['links']
            if extra_data:
                self.payment.extra_data = simplejson.dumps(extra_data)
        self.payment.change_status('waiting')
        self.payment.save()
        raise RedirectNeeded(redirect_to)

    def process_data(self, request):
        extra_data = (simplejson.loads(self.payment.extra_data)
                      if self.payment.extra_data else {})
        success_url = self.payment.get_success_url()
        if not 'token' in request.GET:
            return HttpResponseForbidden('FAILED')
        payer_id = request.GET.get('PayerID')
        if not payer_id:
            if self.payment.status != 'confirmed':
                self.payment.change_status('rejected')
                self.payment.save()
                return redirect(self.payment.get_failure_url())
            else:
                return redirect(success_url)
        response = self.get_payment_execute_response(payer_id)
        response.raise_for_status()
        extra_data['payer_id'] = payer_id
        self.payment.extra_data = simplejson.dumps(extra_data)
        self.payment.change_status('confirmed')
        self.payment.save()
        return redirect(success_url)


class PaypalCardProvider(PaypalProvider):
    '''
    paypal.com credit card payment provider
    '''
    def get_form(self, data=None):
        if self.payment.status == 'waiting':
            self.payment.change_status('input')
        form = PaymentForm(data, provider=self, payment=self.payment)
        if form.is_valid():
            raise RedirectNeeded(self.payment.get_success_url())
        return form

    def get_product_data(self, extra_data):
        data = self.get_transactions_data()
        year = extra_data['expiration'].year
        month = extra_data['expiration'].month
        credit_card = {'number': extra_data['number'],
                       'type': extra_data['type'],
                       'expire_month': month,
                       'expire_year': year}
        if 'cvv2' in extra_data and extra_data['cvv2']:
            credit_card['cvv2'] = extra_data['cvv2']
        data['payer'] = {'payment_method': 'credit_card',
                         'funding_instruments': [{'credit_card': credit_card}]}
        return data

    def process_data(self, request):
        return HttpResponseForbidden('FAILED')



from django import forms
from django.utils.translation import ugettext_lazy as _

from ..forms import CreditCardPaymentForm


class PaymentForm(CreditCardPaymentForm):

    VALID_TYPES = ['visa', 'mastercard', 'discover', 'amex']

    name = forms.CharField(label=_('Name on Credit Card'), max_length=128)

    def clean(self):
        cleaned_data = super(PaymentForm, self).clean()

        if not self.errors:
            if not self.payment.transaction_id:
                request_data = {'type': cleaned_data.get('number').card_type}
                request_data.update(cleaned_data)
                response = self.provider.get_payment_response(cleaned_data)
                data = response.json()
                if response.ok:
                    self.payment.transaction_id = data['id']
                    self.payment.change_status('confirmed')
                else:
                    errors = [error['issue'] for error in data['details']]
                    self._errors['__all__'] = self.error_class(errors)
                    self.payment.change_status('error')
        return cleaned_data



import binascii

from Crypto.Cipher import AES
from django.shortcuts import redirect

from .. import BasicProvider


class SagepayProvider(BasicProvider):
    '''
    sagepay.com payment provider

    vendor:
        vendor name
    encryption_key:
        encryption key
    endpoint:
        gateway URL to post transaction data to
    '''
    _version = '2.23'
    _action = 'https://live.sagepay.com/gateway/service/vspform-register.vsp'

    def __init__(self, *args, **kwargs):
        self._vendor = kwargs.pop('vendor')
        self._enckey = kwargs.pop('encryption_key')
        self._action = kwargs.pop('endpoint', self._action)
        return super(SagepayProvider, self).__init__(*args, **kwargs)

    def _aes_pad(self, crypt):
        padding = ""
        padlength = 16 - (len(crypt) % 16)
        for _i in range(1, padlength + 1):
            padding += chr(padlength)
        return crypt + padding

    def aes_enc(self, data):
        aes = AES.new(self._enckey, AES.MODE_CBC, self._enckey)
        data = self._aes_pad(data.encode('utf-8'))
        enc = aes.encrypt(data)
        enc = "@" + binascii.hexlify(enc)
        return enc

    def aes_dec(self, data):
        data = data.lstrip('@').decode('utf-8')
        aes = AES.new(self._enckey, AES.MODE_CBC, self._enckey)
        dec = binascii.unhexlify(data)
        dec = aes.decrypt(dec)
        return dec

    def get_hidden_fields(self):
        self.payment.save()
        return_url = self.get_return_url()
        data = {
            'VendorTxCode': self.payment.pk,
            'Amount': "%.2f" % (self.payment.total,),
            'Currency': self.payment.currency,
            'SuccessURL': return_url,
            'FailureURL': return_url,
            'Description': "Payment #%s" % (self.payment.pk,),
            'BillingSurname': self.payment.billing_last_name,
            'BillingFirstnames': self.payment.billing_first_name,
            'BillingAddress1': self.payment.billing_address_1,
            'BillingAddress2': self.payment.billing_address_2,
            'BillingCity': self.payment.billing_city,
            'BillingPostCode': self.payment.billing_postcode,
            'BillingCountry': self.payment.billing_country_code,
            'DeliverySurname': self.payment.billing_last_name,
            'DeliveryFirstnames': self.payment.billing_first_name,
            'DeliveryAddress1': self.payment.billing_address_1,
            'DeliveryAddress2': self.payment.billing_address_2,
            'DeliveryCity': self.payment.billing_city,
            'DeliveryPostCode': self.payment.billing_postcode,
            'DeliveryCountry': self.payment.billing_country_code}
        udata = u"&".join(u"%s=%s" % kv for kv in data.items())
        crypt = self.aes_enc(udata)
        return {'VPSProtocol': self._version, 'TxType': 'PAYMENT',
                'Vendor': self._vendor, 'Crypt': crypt}

    def process_data(self, request):
        udata = self.aes_dec(request.GET['crypt'])
        data = {}
        for kv in udata.split('&'):
            k, v = kv.split('=')
            data[k] = v
        success_url = self.payment.get_success_url()
        if self.payment.status == 'waiting':
            # If the payment is not in waiting state, we probably have a page reload.
            # We should neither throw 404 nor alter the payment again in such case.
            if data['Status'] == 'OK':
                self.payment.change_status('confirmed')
                return redirect(success_url)
            else:
                # XXX: We should recognize AUTHENTICATED and REGISTERED in the future.
                self.payment.change_status('rejected')
                return redirect(self.payment.get_failure_url())
        return redirect(success_url)



from django import forms
from django.utils.translation import ugettext_lazy as _

from .fields import CreditCardNumberField, CreditCardExpiryField, CreditCardVerificationField


class PaymentForm(forms.Form):
    '''
    Payment form, suitable for Django templates.

    When displaying the form remeber to use *action* and *method*.
    '''

    #: Form action URL for template use
    action = ''
    #: Form method for template use, either "get" or "post"
    method = 'post'

    def __init__(self, data=None, action=None, method='post', provider=None,
                 payment=None, hidden_inputs=True):
        if hidden_inputs:
            super(PaymentForm, self).__init__(auto_id=False)
            for key, val in data.items():
                widget = forms.widgets.HiddenInput()
                self.fields[key] = forms.CharField(initial=val, widget=widget)
        else:
            super(PaymentForm, self).__init__(data=data)
        self.action = action
        self.method = method
        self.provider = provider
        self.payment = payment


class CreditCardPaymentForm(PaymentForm):

    number = CreditCardNumberField(label=_('Card Number'), max_length=32,
                                   required=True)
    expiration = CreditCardExpiryField()
    cvv2 = CreditCardVerificationField(label=_('CVV2 Security Number'), required=False)

    def __init__(self, *args, **kwargs):
        super(CreditCardPaymentForm, self).__init__(hidden_inputs=False, *args, **kwargs)
        if hasattr(self, 'VALID_TYPES'):
            self.fields['number'].valid_types = self.VALID_TYPES
